#pragma once

#include "parser.hpp"
#include <stdexcept>
#include <string>

namespace console
{

using LoadWarningFn = void(*)(std::string_view warning);
extern void       reset_cvars();
extern void        load_cvars(std::string_view source, LoadWarningFn warning_callback = nullptr);
extern std::string save_cvars();

} // namespace console

namespace console::detail
{

class CVarBase;
extern void register_cvar(CVarBase&);

class CVarBase
{
    friend void register_cvar(CVarBase&);
    friend void ::console::load_cvars(std::string_view, LoadWarningFn);
    friend std::string console::save_cvars();
protected:
    ~CVarBase() noexcept = default;

    virtual std::string serialize()      const          = 0;
    virtual bool deserialize(std::string_view) noexcept = 0;
    virtual bool validate()                    noexcept = 0;

public:
    virtual std::string_view category()  const noexcept = 0;
    virtual std::string_view name()      const noexcept = 0;

    virtual void reset() = 0;
};

struct cvar_attribute_validate {};

} // namespace console::detail

namespace console
{

template <typename T>
class CVar final : detail::CVarBase
{
    using ValidateFn = bool(*)(T const& value);

    std::string_view category_;
    std::string_view name_;
    T default_value_;
    T current_value_;

    ValidateFn validate_ { nullptr };

    static void load_attr() noexcept { }
    template <typename... Attr>
    void load_attr(detail::cvar_attribute_validate, ValidateFn fn, Attr&&... attr)
    {
        if (fn == nullptr) { throw std::invalid_argument{ "nullptr validation function" }; }
        if (validate_ != nullptr) { throw std::logic_error{ "multiple validation functions" }; }
        validate_ = fn;
        if (!(*validate_)(default_value_)) { throw std::logic_error{ "invalid default value" }; }
        load_attr(std::forward<Attr>(attr)...);
    }

    bool validate() noexcept override
    {
        if (validate_ == nullptr ||
            (*validate_)(current_value_))
        {
            return true;
        }
        current_value_ = default_value_;
        return false;
    }
    std::string serialize() const override
    {
        return ConsoleParser<T>::serialize(current_value_);
    }
    bool deserialize(std::string_view in) noexcept override
    {
        return ConsoleParser<T>::deserialize(current_value_, in);
    }
    void reset() override { current_value_ = default_value_; }

    std::string_view category() const noexcept override { return category_; }
    std::string_view name()     const noexcept override { return name_; }

public:
    template <typename... Attr>
    explicit CVar(std::string_view category, std::string_view name,
                  T const& default_value, Attr&&... attr) :
        category_ { category },
        name_     { name },
        default_value_ { default_value },
        current_value_ { default_value }
    {
        load_attr(std::forward<Attr>(attr)...);
        detail::register_cvar(*this);
    }

    T const& operator*() const noexcept { return current_value_; }
    void set(T value)
    {
        if (validate_ != nullptr && !(*validate_)(value))
        {
            throw std::invalid_argument{ "invalid cvar" };
        }
        current_value_ = value;
    }
};

} // namespace console

#define CVAR_DECLARE(TYPE, VAR_NAME) extern ::console::CVar<TYPE> VAR_NAME
#define CVAR_DEFINE(TYPE, VAR_NAME, CATEGORY, NAME, DEFAULT_VALUE, ...) \
    ::console::CVar<TYPE> VAR_NAME{ #CATEGORY, #NAME, DEFAULT_VALUE, __VA_ARGS__ }

#define CVAR_VALIDATE(FN) ::console::detail::cvar_attribute_validate{}, (FN)
