#include <console.hpp>
#include <map>
#include <unordered_set>

namespace console
{
    
namespace
{
    
using DataMap = std::multimap<std::string_view, detail::CVarBase*>;

// Trick to avoid static initialization order issues.
// The data() function will create a DataMap and then
// guarantee that it's initialized. It also assigns
// the data_ variable. Any code not running during static
// initialization can then directly access the DataMap.
DataMap& data()
{
    static DataMap map;
    return map;
}
DataMap const& data_ = data();

// Removes the first line from a source string_view
// and returns a view to the removed line.
std::string_view extract_line(std::string_view& source)
{
    if (source.empty()) { return {}; }
    // Iterate till end or new line character, skipping empty lines
    size_t start = 0;
    size_t i = 0;
    for (; i < source.size(); ++i)
    {
        auto const c = source[i];
        if (c != '\r' && c != '\n') { continue; }
        // Ignore newlines at the start of the string
        if (start == i)
        {
            ++start;
            continue;
        }
        break;
    }

    // start == inclusive start of the extracted string
    // i     == exclusive end   of the extracted string

    // Iterate till first non-newline character or the end of the
    // of the string, this will be what the source is trimmed to.
    size_t j = i;
    for (; j < source.size(); ++j)
    {
        auto const c = source[j];
        if (c != '\r' && c != '\n') { break; }
    }

    auto result = source.substr(start, i - start);
    source = source.substr(j);
    return result;
}

std::string_view trim_start_whitespace(std::string_view const in)
{
    for (size_t i = 0; i < in.size(); ++i)
    {
        auto const c = in[i];
        if (c >= 0x20)
        {
            return in.substr(i);
        }
    }
    return {};
}

std::string_view trim_end_whitespace(std::string_view const in)
{
    for (size_t i = in.size(); i != 0; --i)
    {
        auto const c = in[i - 1];
        if (c >= 0x20)
        {
            return in.substr(0, i);
        }
    }
    return {};
}

std::string_view trim_whitespace(std::string_view const in)
{
    return trim_start_whitespace(trim_end_whitespace(in));
}

struct StringViewPairHash
{
    using Type = std::pair<std::string_view, std::string_view>;
    size_t operator()(Type const& t) const noexcept
    {
        constexpr auto hasher = std::hash<std::string_view>{};
        return hasher(t.first) ^ hasher(t.second);
    }
};
using CVarsSet = std::unordered_set<std::pair<std::string_view, std::string_view>, StringViewPairHash>;

template <typename... Args>
void trigger_warning(LoadWarningFn warning_fn, Args&&... params)
{
    if (warning_fn == nullptr) { return; }
    std::stringstream ss;
    // Invoke operator<< for each parameter
    auto _ = { (static_cast<void>(ss << std::forward<Args>(params)), 0)... };
    (*warning_fn)(ss.str());
}

} // namespace

namespace detail // console::detail
{
    
// WARNING: Runs during static initialization
void register_cvar(CVarBase& cvar)
{
    auto& map = data();
    auto const name     = cvar.name();
    auto const category = cvar.category();

    // Check for duplicate cvars, by iterating over each cvar in the category.
    auto const cvars_in_category = map.equal_range(category);
    for (auto it = cvars_in_category.first; it != cvars_in_category.second; ++it)
    {
        if (name == it->second->name())
        {
            std::stringstream ss;
            ss << "duplicate cvar [" << category << "] " << name;
            throw std::logic_error{ ss.str() };
        }
    }

    // No duplicate entries, so insert and done
    map.insert({ category, &cvar });
}

} // namespace console::detail

void reset_cvars()
{
    for (auto& pair : data_)
    {
        pair.second->reset();
    }
}

void load_cvars(std::string_view source, LoadWarningFn const warning_callback)
{
    if (source.empty()) { return; }

    std::string_view current_category{};
    CVarsSet loaded_cvars;
    do
    {
        auto line = trim_start_whitespace(extract_line(source));
        // Ignore empty lines
        if (line.empty()) { continue; }

        // Ignore comments
        if (line[0] == '#') { continue; }

        // Parse category
        if (line[0] == '[')
        {
            line = trim_end_whitespace(line);
            if (line.back() != ']')
            {
                trigger_warning(warning_callback,
                    "parsed '[' but line does not end with ']' in line: ", line);
                continue;
            }
            auto const category = trim_whitespace(line.substr(1, line.size() - 2));
            if (category.empty())
            {
                trigger_warning(warning_callback, "empty category");
                continue;
            }
            current_category = category;
            continue;
        }

        // Parse cvar key
        auto index = line.find_first_of('=');
        if (index == std::string_view::npos)
        {
            trigger_warning(warning_callback, "could not find equal sign in cvar in line: ", line);
            continue;
        }
        auto cvar_name = trim_end_whitespace(line.substr(0, index));
        if (cvar_name.empty())
        {
            trigger_warning(warning_callback, "empty cvar name in line: ", line);
            continue;
        }

        if (current_category.empty())
        {
            trigger_warning(warning_callback, "cvar before category in line: ", line);
            continue;
        }

        // Prevent duplicate category/name combinations
        if (!loaded_cvars.insert({ current_category, cvar_name }).second)
        {
            trigger_warning(warning_callback,
                "cvar [", current_category, "] ", cvar_name,
                " is defined twice, first definition used");
            continue;
        }

        // Search for the cvar category & name
        auto const cvars_in_category = data_.equal_range(current_category);
        bool found = false;
        for (auto it = cvars_in_category.first; it != cvars_in_category.second; ++it)
        {
            auto& cvar = *it->second;
            if (cvar.name() == cvar_name)
            {
                found = true;
                // Try to load the cvar
                auto const cvar_value = line.substr(index + 1);
                if (!cvar.deserialize(cvar_value))
                {
                    trigger_warning(warning_callback,
                        "cvar [", current_category, "] ", cvar_name, " could not be parsed");
                }
                else if (!cvar.validate())
                {
                    trigger_warning(warning_callback,
                        "cvar [", current_category, "] ", cvar_name, " invalid value");
                }
                break;
            }
        }
        if (!found)
        {
            trigger_warning(warning_callback,
                "cvar [", current_category, "] ", cvar_name, " does not exist");
        }
    } while (!source.empty());
}

std::string save_cvars()
{
    std::stringstream ss;
    auto const begin = data_.begin();
    auto const end   = data_.end();
    if (begin == end)
    {
        return "";
    }
    std::string_view current_category{};
    for (auto it = begin; it != end; ++it)
    {
        if (it->first != current_category)
        {
            if (!current_category.empty())
            {
                ss << "\n";
            }
            current_category = it->first;
            ss << "[" << current_category << "]\n";
        }
        auto& cvar = *it->second;
        std::string const value = cvar.serialize();
        ss << cvar.name() << "=" << value << "\n";
    }
    return ss.str();
}

} // namespace console
