#include <framework/game.hpp>
#include <framework/surface.hpp>
#include <console.hpp>
#include <sstream>
#include <iomanip>
#include <iostream>
#include <fstream>
#include <utility>

using namespace fw;

namespace
{

bool validate_size(const i32& s) noexcept { return s >= 128; }

CVAR_DEFINE(i32, window_width,  Window, Width,  128, CVAR_VALIDATE(validate_size));
CVAR_DEFINE(i32, window_height, Window, Height, 128, CVAR_VALIDATE(validate_size));
CVAR_DEFINE(i32, window_scale_power, Window, ScalePower, 3,
    CVAR_VALIDATE([](auto value) noexcept { return value >= 0 && value <= 3; })
);

CVAR_DEFINE(Color, background_color, Colors, BackgroundColor, { 0, 0, 128 });
CVAR_DEFINE(Color, foreground_color, Colors, ForegroundColor, { 255, 0, 0 });

void save_cvars_to_file()
{
    std::ofstream cvar_file{ "cvars" };
    cvar_file << console::save_cvars();
}

void load_cvars_from_file()
{
    console::reset_cvars();
    std::ifstream cvar_file{ "cvars" };
    std::stringstream cvar_str;
    cvar_str << cvar_file.rdbuf();
    console::load_cvars(cvar_str.str(), [](std::string_view warning)
    {
        std::cout << "warning loading cvars: " << warning << "\n";
    });
}

} // namespace

class CvarGame : public Game
{
public:
    CvarGame() :
        Game{ "Cvar example", *window_width, *window_height, *window_scale_power }
    {
    }

    void update(Surface& screen, TimeSpan const /*delta_time*/) override
    {
        screen.clear(*background_color);
        screen.draw_text("Hello World!", 10, 10, *foreground_color);
    }

    void on_key_down(Key const key, Modifiers /*modifiers*/, bool /*repeat*/) override
    {
        switch (key)
        {
        case Key::Escape:
            exit();
            return;
        case Key::F1:
            console::reset_cvars();
            save_cvars_to_file();
            break;
        case Key::F2:
            load_cvars_from_file();
            break;
        default: return;
        }
        resize(*window_width, *window_height, *window_scale_power);
    }
};

int main()
{
    try
    {
        load_cvars_from_file();
        Game::run(CvarGame{});
        save_cvars_to_file();
    }
    catch (std::exception const& ex)
    {
        std::cerr << "Error while running game:\n" << ex.what() << "\n";
        return -1;
    }
    return 0;
}
