// #include "private/particle_framework.hpp"
#include "private/particle_system.hpp"
#include "private/particle_random.hpp"

using namespace std::string_view_literals;

namespace particles
{

struct ExampleParticle
{
    float px, py;
    float vx, vy;

    float life_rate_     = 0.f; // 1 / total_lifetime
    float life_progress_ = 0.f;

    void update(float delta_seconds) noexcept
    {
        life_progress_ += life_rate_ * delta_seconds;

        px += vx * delta_seconds;
        py += vy * delta_seconds;
    }

    auto is_alive() const noexcept -> bool
    { return life_progress_ < 1.f; }
    auto position() const noexcept -> std::array<float, 2>
    { return { px, py }; }
    void position(float x, float y) noexcept
    {
        px = x;
        py = y;
    }
};

struct ExampleParticleType
{
    float life_min, life_max;
    float pos_min, pos_max;
    float vel_min, vel_max;

    using Particle = ExampleParticle;
    Particle create(RandomEngine& engine, float x, float y) const
    {
        std::uniform_real_distribution<float> rnd_life{ life_min, life_max };
        std::uniform_real_distribution<float> rnd_pos {  pos_min,  pos_max };
        std::uniform_real_distribution<float> rnd_vel {  vel_min,  vel_max };
        return                                                                            
        {
            rnd_pos(engine) + x, rnd_pos(engine) + y,
            rnd_vel(engine)    , rnd_vel(engine)    ,
            1 / rnd_life(engine)
        };
    }
};

constexpr auto example_slow   = ExampleParticleType{ 0.6f, 1.5f,  -2.f,  2.f, -30.f, 30.f };
constexpr auto example_medium = ExampleParticleType{ 0.5f, 1.2f,  -4.f,  4.f, -50.f, 50.f };
constexpr auto example_fast   = ExampleParticleType{ 0.4f, 0.8f, -20.f, 20.f, -70.f, 70.f };
constexpr auto example_system = CreateParticleSystem("example"sv,
    EmitOnDeath(example_slow, 1, EmitOnDeath(example_slow, 3, EmitContinuously(example_medium, 5.f))),
    EmitContinuously(example_fast, 10.f)
);

// constexpr auto fire = CreateParticleType(
//     Lifetime(Random{ 0.2f, 0.8f }),
//     Position(Random{ -2.f, 2.f }),
//     Velocity(Random{ -4.f, 4.f }, Random{ 0.f, 50.f })
// );

// constexpr auto sparks = CreateParticleType(
//     Lifetime(Random{ 0.2f, 0.4f }),
//     Position(Random{ -4.f, 4.f }),
//     Velocity(Random{ -200.f, 200.f }),
//     Acceleration(Constant{ 0.5f }, Constant{ -1500.f }),
//     Drag(Constant{ 3.f })
// );

// constexpr auto swarm = CreateParticleType(
//     Lifetime(Constant{ 1.f }),
//     Position(Random{ -7.5f, 7.5f }),
//     Acceleration(Polar{ 0.f, 15.f }),
//     Drag(Constant{ 0.5f })
// );

// constexpr auto rain = CreateParticleType(
//     Lifetime(Constant{ 1.f }),
//     Position(Random{ -20.f, 0.f }, Constant { 10.f }),
//     Velocity(Random{ 75.f, 85.f }, Random{ -410.f, -390.f })
// );

// constexpr auto firework_rocket = CreateParticleType(
//     Lifetime(Random{ 0.4f, 0.5f }),
//     Position(Constant{ 0.f }),
//     Velocity(Polar{ 250.f, 350.f, pi * 0.45f, pi * 0.55f }),
//     Acceleration(Constant{ 0.f }, Constant{ -300.f })
// );
// constexpr auto firework_explosion = CreateParticleType(
//     Lifetime(Random{ 0.6f, 0.9f }),
//     Velocity(Polar{ 150.f, 200.f }),
//     Acceleration(Constant{ 0.f }, Constant{ -100.f }),
//     Drag(Constant{ 2.f })
// );

// constexpr auto fire_system     = CreateParticleSystem("fire"sv,   EmitContinuously(fire,   20.f));
// constexpr auto sparks_system   = CreateParticleSystem("sparks"sv, EmitContinuously(sparks, 120.f));
// constexpr auto swarm_system    = CreateParticleSystem("swarm"sv,  EmitContinuously(swarm,  20.f));
// constexpr auto rain_system     = CreateParticleSystem("rain"sv,   EmitContinuously(rain,   40.f));
// constexpr auto firework_system = CreateParticleSystem("firework"sv,
//     EmitOnDeath(firework_explosion, 100, EmitContinuously(firework_rocket, 0.75f))
// );

ParticleSimulation* create_simulation() noexcept
{
    try
    {
        auto ps = new ParticleSimulationImpl;
        ps->register_particle_system(example_system);
        // ps->register_particle_system(fire_system);
        // ps->register_particle_system(sparks_system);
        // ps->register_particle_system(swarm_system);
        // ps->register_particle_system(rain_system);
        // ps->register_particle_system(firework_system);
        return ps;
    }
    catch (std::bad_alloc const&)
    {
        return nullptr;
    }
}

void destroy_simulation(ParticleSimulation* ptr) noexcept
{
    delete ptr;
}

} // namespace particles
