#pragma once

#include <particles/particle_simulation.hpp>
#include "particle_random.hpp"
#include <stdexcept>
#include <unordered_map>
#include <memory>
#include <string_view>

namespace particles
{

class ParticleSystem
{
public:
    virtual void update(fw::Game::TimeSpan delta_time, RandomEngine& random) = 0;
    virtual void render(fw::Surface& screen) const noexcept = 0;

    virtual ~ParticleSystem() noexcept = default;
};

class ParticleSimulationImpl : public ParticleSimulation
{
    // For type erasure
    using SpawnFn = std::unique_ptr<ParticleSystem>(*)(void const*, fw::f32, fw::f32);
    struct Descr
    {
        void const* context;
        SpawnFn spawner;
    };

    std::unordered_map<std::string_view, Descr> descriptions_;
    std::vector<std::string_view> names_;
    std::vector<std::unique_ptr<ParticleSystem>> systems_;
    RandomEngine rnd_{};

public:
    template <typename T>
    void register_particle_system(T const& particle_system)
    {
        // Uses type erasure through function pointers
        SpawnFn spawner = [](void const* object, fw::f32 x, fw::f32 y)
        {
            return static_cast<T const*>(object)->spawn(x, y);
        };
        if (!descriptions_.insert({ particle_system.name(), { static_cast<void const*>(&particle_system), spawner } }).second)
        {
            throw std::invalid_argument{ "duplicate particle system name" };
        }
        names_.push_back(particle_system.name());
    }

    auto get_particle_types() const noexcept
        -> std::vector<std::string_view> const& override
    {
        return names_;
    }
    void spawn_particle_system(std::string_view type, fw::f32 x, fw::f32 y) override
    {
        auto it = descriptions_.find(type);
        if (it == descriptions_.end())
        {
            throw std::invalid_argument{ "no such ParticleSystem" };
        }
        systems_.push_back(it->second.spawner(it->second.context, x, y));
    }
    void reset() noexcept override
    {
        systems_.clear();
    }
    void update(fw::Game::TimeSpan delta_time) override
    {
        // C++17 Structured binding
        for (auto& system : systems_)
        {
            system->update(delta_time, rnd_);
        }
    }
    void render(fw::Surface& screen) noexcept override
    {
        // C++17 Structured binding
        for (auto& system : systems_)
        {
            system->render(screen);
        }
    }
};

} // namespace particles
