#pragma once

#include "particle_simulation.hpp"
#include <framework/surface.hpp>
#include <utility>
#include <tuple>

namespace particles::detail
{

struct SpawnInfo
{
    SpawnInfo() = default;
    RandomEngine& random;
    float x;
    float y;
};

template <typename ParticleType>
struct EmitterBase
{
    using Particle = typename ParticleType::Particle;
    std::vector<Particle> particles_;

    static void default_on_death(Particle const&) noexcept {}
    template <typename OnDeath = decltype(default_on_death)>
    void update([[maybe_unused]] float const total_time, float const delta_time,
        [[maybe_unused]] SpawnInfo spawn_info, OnDeath on_death = default_on_death)
    {
        for (size_t i = particles_.size(); i != 0;)
        {
            --i;
            particles_[i].update(delta_time);
            if (!particles_[i].is_alive())
            {
                on_death(particles_[i]);
                using std::swap;
                swap(particles_[i], particles_[particles_.size() - 1]);
                particles_.pop_back();
            }
        }
    }

    void spawn_particle(ParticleType type, SpawnInfo spawn_info)
    {
        particles_.push_back(type.create(spawn_info.random, spawn_info.x, spawn_info.y));
    }

    void render(fw::Surface& screen) const noexcept
    {
        for (auto& particle : particles_)
        {
            // C++17 Structured binding
            auto const[fx, fy] = particle.position();
            auto const x = static_cast<fw::i32>(fx);
            auto const y = static_cast<fw::i32>(fy);
            if (x >= 0 && x < screen.width() &&
                y >= 0 && y < screen.height())
            {
                screen[x][y] = 0xffffffff;
            }
        }
    }
};

template <typename T>
struct EmitterContinuous : EmitterBase<T>
{
    using Base = EmitterBase<T>;
    T type_;
    fw::f32 spawn_rate_;
    fw::i32 spawn_count_ = 0;

    EmitterContinuous(T type, fw::f32 spawn_rate) :
        type_{ type },
        spawn_rate_{ spawn_rate }
    { }

    template <typename OnDeath = decltype(Base::default_on_death)>
    void update(float const total_time, float const delta_time,
        SpawnInfo spawn_info, OnDeath on_death = Base::default_on_death)
    {
        Base::update(total_time, delta_time, spawn_info, on_death);
        auto const spawn_target = static_cast<fw::i32>(total_time * spawn_rate_);
        while (spawn_count_ < spawn_target)
        {
            ++spawn_count_;
            Base::spawn_particle(type_, spawn_info);
        }
    }
};

template <typename T>
struct EmitterDescrContinuous
{
    using Emitter = EmitterContinuous<T>;

    T particle_type;
    float spawn_rate; // in seconds between each spawn

    Emitter spawn() const noexcept
    {
        return { particle_type, spawn_rate };
    }
};

template <typename T, typename ParentEmitter>
struct EmitterOnDeath : EmitterBase<T>
{
    using Base = EmitterBase<T>;
    T type_;
    fw::i32 spawn_count_;
    ParentEmitter parent_emitter_;

    EmitterOnDeath(T type, fw::i32 spawn_count, ParentEmitter parent_emitter) :
        type_{ type },
        spawn_count_{ spawn_count },
        parent_emitter_{ parent_emitter }
    { }

    template <typename OnDeath = decltype(Base::default_on_death)>
    void update(float const total_time, float const delta_time,
        SpawnInfo spawn_info, OnDeath on_death = Base::default_on_death)
    {
        Base::update(total_time, delta_time, spawn_info, on_death);
        parent_emitter_.update(total_time, delta_time, spawn_info, [&](auto const& particle)
        {
            for (auto i = 0; i < spawn_count_; ++i)
            {
                auto[px, py] = particle.position();
                Base::spawn_particle(type_, { spawn_info.random, px, py });
            }
        });
    }

    void render(fw::Surface& screen) const noexcept
    {
        parent_emitter_.render(screen);
        Base::render(screen);
    }
};

template <typename T, typename ParentEmitter>
struct EmitterDescrOnDeath
{
    T particle_type;
    fw::i32 spawn_count;
    ParentEmitter parent;

    using Emitter = EmitterOnDeath<T, typename ParentEmitter::Emitter>;

    Emitter spawn() const noexcept
    {
        return { particle_type, spawn_count, parent.spawn() };
    }
};

template <typename... Emitters>
struct ParticleSystemImpl : ParticleSystem
{
    using Data = std::tuple<typename Emitters::Emitter...>;
    Data emitters_;
    fw::f32 x_;
    fw::f32 y_;

    float total_time_ = 0.f;

    ParticleSystemImpl(fw::f32 x, fw::f32 y, Data emitters) :
        x_{ x },
        y_{ y },
        emitters_{ emitters }
    {}

    void update(fw::Game::TimeSpan delta_time, RandomEngine& random) override
    {
        total_time_ += delta_time.count();
        update_impl(delta_time.count(), random, std::make_index_sequence<sizeof...(Emitters)>{});
    }
    void render(fw::Surface& screen) const noexcept override
    {
        render_impl(screen, std::make_index_sequence<sizeof...(Emitters)>{});
    }

private:
    template <size_t... Indices>
    void update_impl(float delta_time, RandomEngine& random, std::index_sequence<Indices...>)
    {
        auto _ = { (
            std::get<Indices>(emitters_).update(total_time_, delta_time, { random, x_, y_ })
            , 0)... };
    }

    template <size_t... Indices>
    void render_impl(fw::Surface& screen, std::index_sequence<Indices...>) const noexcept
    {
        auto _ = { (
            std::get<Indices>(emitters_).render(screen)
            , 0)... };
    }
};

template <typename... Emitters>
struct ParticleSystemDescr
{
    std::string_view name_;
    std::tuple<Emitters...> emitters;

    auto name() const noexcept -> std::string_view { return name_; }
    auto spawn(fw::f32 x, fw::f32 y) const -> std::unique_ptr<ParticleSystem>
    {
        return spawn_impl(x, y, std::make_index_sequence<sizeof...(Emitters)>{});
    }

    template <size_t... Indices>
    auto spawn_impl(fw::f32 x, fw::f32 y, std::index_sequence<Indices...>) const
        -> std::unique_ptr<ParticleSystem>
    {
        return std::make_unique<ParticleSystemImpl<Emitters...>>(
            x,
            y,
            std::tuple<typename Emitters::Emitter...>{ std::get<Indices>(emitters).spawn()... }
        );
    }
};

} // namespace particles::detail

namespace particles
{

template <typename... Emitters>
constexpr auto CreateParticleSystem(std::string_view name, Emitters... emitters)
{
    return detail::ParticleSystemDescr<Emitters...>{ name, std::tuple<Emitters...>{ emitters... } };
}

template <typename T>
constexpr auto EmitContinuously(T particle_type, float spawn_rate)
    -> detail::EmitterDescrContinuous<T>
{
    return { particle_type, spawn_rate };
}

template <typename T, typename Emitter>
constexpr auto EmitOnDeath(T particle_type, fw::i32 spawn_count, Emitter parent)
    -> detail::EmitterDescrOnDeath<T, Emitter>
{
    return { particle_type, spawn_count, parent };
}

} // namespace particles
