#pragma once

#include <type_traits>
#include <stddef.h> // size_t

namespace particles::detail
{

/*
 * Represent an arbitrary list of types.
 */
template <typename... Types>
struct TypeList {};

/*
 * ConcatTypeList<typename TA, typename TB>
 * Concatenates the types in the latter typelist to the
 * ones in the former.
 * 
 * Example:
 * ConcatTypeList<
 *     TypeList<bool, float, int>,
 *     TypeList<char, unsigned>
 * > === TypeList<bool, float, int, char, unsigned>
 */
template <typename... TA, typename... TB>
constexpr auto concat_type_list(TypeList<TA...>, TypeList<TB...>)
-> TypeList<TA..., TB...> {
    return {};
}

template <typename TA, typename TB>
using ConcatTypeList = decltype(concat_type_list(TA{}, TB{}));

/*
 * IndexTypes gets the Nth element from a variadic pack.
 * 
 * Examples:
 * IndexTypes<0, int, float, bool, short> === int
 * IndexTypes<2, int, float, bool, short> === bool
 */
template <size_t Index, typename First, typename... Rest>
struct IndexTypesImpl
{
    static_assert(Index <= sizeof...(Rest), "index out of range");
    using Type = typename IndexTypesImpl<
        Index - 1,
        Rest...
    >::Type;
};
template <typename First, typename... Rest>
struct IndexTypesImpl<0, First, Rest...>
{
    using Type = First;
};

template <size_t Index, typename... Types>
using IndexTypes = typename IndexTypesImpl<Index, Types...>::Type;

/*
 * ContainsType and ListContainsType check if a
 * a type is contained in a list of types.
 * 
 * The first template parameter is the type being
 * searched for.
 * 
 * Examples:
 * ContainsType<short, int, float, double> == false
 * ContainsType<float, int, float, double> == true
 * ListContainsType<int, TypeList<float, int>> == true
 */
template <typename T, typename... Types>
constexpr bool ContainsType = false;
template <typename T, typename... Rest>
constexpr bool ContainsType<T, T, Rest...> = true;
template <typename T, typename J, typename... Rest>
constexpr bool ContainsType<T, J, Rest...> = ContainsType<T, Rest...>;

template <typename T, typename TypeList>
constexpr bool ListContainsType = false;
template <typename T, typename... Types>
constexpr bool ListContainsType<T, TypeList<Types...>> = ContainsType<T, Types...>;

/*
 * Appends a type if it's not already in the list.
 */
template <typename T, typename... Types>
using AppendTypeUnique = std::conditional_t<
    ContainsType<T, Types...>,
    /*?*/TypeList<Types...>,
    /*:*/TypeList<Types..., T>>;

/*
 * Filters all non-unique elements from a list of types.
 * 
 * Example:
 * UniqueTypeList<int, float, long, int, double, short, short>
 * === TypeList<int, float, long, double, short>
 */
template <typename... Results>
constexpr auto unique_type_list(TypeList<Results...>, TypeList<>)
    -> TypeList<Results...> { return {}; }
template <typename... Results, typename First, typename... Inputs>
constexpr auto unique_type_list(TypeList<Results...>, TypeList<First, Inputs...>)
    -> decltype(unique_type_list(AppendTypeUnique<First, Results...>{}, TypeList<Inputs...>{}))
{ return {}; }

template <typename... Types>
using UniqueTypeList = decltype(unique_type_list(TypeList<>{}, TypeList<Types...>{}));

} // particles::detail
