#pragma once
#include "types.hpp"
#include <array>
#include <string_view>
#include <optional>

namespace fw
{
    
struct Color
{
    u32 value{ 0 };

    constexpr Color() noexcept = default;
    constexpr Color(u32 const value) noexcept :
        value{ value }
    {}

    constexpr Color(u8 r, u8 g, u8 b, u8 a = 255) noexcept :
        value{ u32{b} << 24 | u32{g} << 16 | u32{r} << 8 | u32{a} }
    {}

    constexpr u8 r() const noexcept { return static_cast<u8>(value >> 8); }
    constexpr u8 g() const noexcept { return static_cast<u8>(value >> 16); }
    constexpr u8 b() const noexcept { return static_cast<u8>(value >> 24); }
    constexpr u8 a() const noexcept { return static_cast<u8>(value); }

    constexpr operator u32() const noexcept { return value; }

    constexpr static std::optional<Color> from_hex_string(std::string_view) noexcept;
    constexpr std::array<char, 10> to_hex_string() const noexcept;
};

constexpr bool operator==(Color const& lhs, Color const& rhs) { return lhs.value == rhs.value; }
constexpr bool operator!=(Color const& lhs, Color const& rhs) { return lhs.value != rhs.value; }

constexpr std::optional<Color> Color::from_hex_string(std::string_view str) noexcept
{
    auto const index = str.find_first_of('#');
    if (index == std::string_view::npos) { return {}; }
    if (str.size() - index < 9) { return {}; }

    auto const hex = str.substr(index + 1, 8);
    u32 value = 0;

    for (size_t i = 0; i < 8; ++i)
    {
        auto const c = hex[i];
        u32 cv;
        if (c >= '0' && c <= '9')
        {
            cv = c - '0';
        }
        else if (c >= 'a' && c <= 'f')
        {
            cv = c - 'a' + 0xa;
        }
        else if (c >= 'A' && c <= 'F')
        {
            cv = c - 'A' + 0xa;
        }
        else { return {}; }
        value |= cv << (i * 4 + 4 - (i % 2) * 8);
    }
    return Color{ value };
}

constexpr std::array<char, 10> Color::to_hex_string() const noexcept
{
    std::array<char, 10> data{};
    data[0] = '#';
    data[9] = '\0';
    auto write_hex = [](char& to, i32 value) noexcept
    {
        value &= 0xf;
        to = static_cast<char>(value >= 0xa
            ? value - 0xa + 'a'
            : value + '0');
    };
    for (size_t i = 0; i < 4; ++i)
    {
        write_hex(data[i * 2 + 2], value >> (i * 8));
        write_hex(data[i * 2 + 1], value >> (i * 8 + 4));
    }
    return data;
}

} // namespace fw
