#pragma once

#include "types.hpp"
#include "color.hpp"
#include <memory>
#include <string_view>

namespace fw
{
    
class Surface;
namespace detail { class SurfaceIndexer; class ConstSurfaceIndexer; }

struct BlendAdd;

class Surface
{
public:
    // Default operations
    Surface() noexcept = default;
    ~Surface() noexcept = default;

    Surface           (Surface const&);
    Surface& operator=(Surface const&);
    Surface           (Surface&&) noexcept;
    Surface& operator=(Surface&&) noexcept;

    explicit Surface(i32 width, i32 height, Color clear_color = {0});

    void swap(Surface& rhs) noexcept;

    explicit operator bool() const noexcept { return buffer_ != nullptr; }

    Color const* data()      const noexcept { return buffer_.get(); }
    Color      * data()            noexcept { return buffer_.get(); }
    Color const* operator*() const noexcept { return buffer_.get(); }
    Color      * operator*()       noexcept { return buffer_.get(); }

    i32 width()  const noexcept { return width_; }
    i32 height() const noexcept { return height_; }

    // Range accessors
    Color  get(i32 x, i32 y) const;
    Color& get(i32 x, i32 y);

    // Unchecked accessors (careful for buffer overflow)
    detail::ConstSurfaceIndexer operator[](i32 x) const noexcept;
    detail::SurfaceIndexer      operator[](i32 x)       noexcept;

    void clear(Color color) noexcept;

    void draw_surface(Surface const& surface, i32 x, i32 y) noexcept;
    template <typename BlendFn = BlendAdd>
    void draw_surface(Surface const& surface, i32 x, i32 y, BlendFn blend_fn = BlendFn{}) noexcept(std::is_nothrow_invocable_v<BlendFn, Color, Color>);
    void draw_text(std::string_view text, i32 x, i32 y, Color color) noexcept;
    void draw_line(i32 x1, i32 y1, i32 x2, i32 y2, Color color) noexcept;
    void draw_box (i32 x1, i32 y1, i32 x2, i32 y2, Color color) noexcept;
    void draw_bar (i32 x1, i32 y1, i32 x2, i32 y2, Color color) noexcept;
    
    template <typename FillFn>
    std::enable_if_t<std::is_invocable_r_v<Color, FillFn, i32, i32>>
        fill(FillFn fill_fn = FillFn{})
        noexcept(std::is_nothrow_invocable_v<FillFn, i32, i32>);
    template <typename FillFn>
    std::enable_if_t<std::is_invocable_r_v<Color, FillFn, i32, i32, Color>>
        fill(FillFn fill_fn = FillFn{})
        noexcept(std::is_nothrow_invocable_r_v<Color, FillFn, i32, i32, Color>);

private:
    i32 width_  = 0;
    i32 height_ = 0;
    std::unique_ptr<Color[]> buffer_ = nullptr;
};

inline void swap(Surface& lhs, Surface& rhs) noexcept
{
    lhs.swap(rhs);
}

} // namespace fw

#include "surface.inl"
