#pragma once
#include "surface.hpp"
#include <algorithm>
#include <cassert>

namespace fw
{

namespace detail
{

class SurfaceIndexer
{
    friend class Surface;
    Surface* s_;
    i32 x_;
    FW_FORCE_INLINE SurfaceIndexer(Surface* s, i32 const x) : s_{ s }, x_{ x } { }

public:
    FW_FORCE_INLINE Color& operator[](i32 const y) const noexcept
    {
        return (**s_)[y * s_->width() + x_];
    }
};
class ConstSurfaceIndexer
{
    friend class Surface;
    Surface const* s_;
    i32 x_;
    FW_FORCE_INLINE ConstSurfaceIndexer(Surface const* s, i32 const x) : s_{ s }, x_{ x } { }

public:
    FW_FORCE_INLINE Color operator[](i32 const y) const noexcept
    {
        return (**s_)[y * s_->width() + x_];
    }
};

} // namespace

inline FW_FORCE_INLINE detail::SurfaceIndexer Surface::operator[](i32 const x) noexcept
{
    assert(*this);
    return { this, x };
}
inline FW_FORCE_INLINE detail::ConstSurfaceIndexer Surface::operator[](i32 const x) const noexcept
{
    assert(*this);
    return { this, x };
}

struct BlendAdd
{
    FW_FORCE_INLINE Color operator()(Color const src, Color const dst) const noexcept
    {
        auto const r = static_cast<u8>(std::min(255, src.r() + dst.r()));
        auto const g = static_cast<u8>(std::min(255, src.g() + dst.g()));
        auto const b = static_cast<u8>(std::min(255, src.b() + dst.b()));
        auto const a = static_cast<u8>(std::min(255, src.a() + dst.a()));
        return { r, g, b, a };
    }
};

struct BlendOverwrite
{
    FW_FORCE_INLINE Color operator()(Color const src, Color const /*dst*/) const noexcept
    {
        return src;
    }
};

template <typename BlendFn>
void Surface::draw_surface(Surface const& surface, i32 const x, i32 const y, BlendFn blend_fn) noexcept(std::is_nothrow_invocable_v<BlendFn, Color, Color>)
{
    assert(*this);
    auto x1 = x;
    auto x2 = x + surface.width_ - 1;
    auto y1 = y;
    auto y2 = y + surface.height_ - 1;

    if (x1 >= width_  || x2 < 0 ||
        y1 >= height_ || y2 < 0)
    {
        return;
    }

    i32 sdx = 0, sdy = 0;
    if (x1 < 0) { sdx = -x1; x1 = 0; }
    if (y1 < 0) { sdy = -y1; y1 = 0; }
    x2 = std::min(x2, width_ - 1);
    y2 = std::min(y2, height_ - 1);

    for (auto dy = y1, sy = sdy; dy <= y2; ++dy, ++sy)
    {
        for (auto dx = x1, sx = sdx; dx <= x2; ++dx, ++sx)
        {
            get(dx, dy) = blend_fn(surface[sx][sy], get(dx, dy));
        }
    }
}

inline void Surface::draw_surface(Surface const& surface, i32 const x, i32 const y) noexcept
{
    draw_surface<BlendOverwrite>(surface, x, y);
}

template <typename FillFn>
std::enable_if_t<std::is_invocable_r_v<Color, FillFn, i32, i32>>
    Surface::fill(FillFn fill_fn)
    noexcept(std::is_nothrow_invocable_v<FillFn, i32, i32>)
{
    for (i32 y = 0; y < width_; ++y)
    {
        for (i32 x = 0; x < height_; ++x)
        {
            (*this)[x][y] = fill_fn(x, y);
        }
    }
}

template <typename FillFn>
std::enable_if_t<std::is_invocable_r_v<Color, FillFn, i32, i32, Color>>
    Surface::fill(FillFn fill_fn)
    noexcept(std::is_nothrow_invocable_r_v<Color, FillFn, i32, i32, Color>)
{
    for (i32 y = 0; y < width_; ++y)
    {
        for (i32 x = 0; x < height_; ++x)
        {
            (*this)[x][y] = fill_fn(x, y, (*this)[x][y]);
        }
    }
}

} // namespace fw
