#include <framework/game.hpp>
#include <framework/surface.hpp>
#include "private/render_incl.hpp"
#include "private/shader.hpp"
#include "private/shader_program.hpp"
#include "private/texture.hpp"
#include <stdexcept>
#include <iostream>

namespace fw
{

namespace
{

constexpr char const* vert_shader = R"shader(
#version 330 core
layout(location = 0) in vec2 v_pos;
out vec2 f_uv;
void main()
{
    gl_Position = vec4(v_pos, 0.f, 1.f);
    f_uv = v_pos * 0.5f + 0.5f;
}
)shader";
constexpr char const* frag_shader = R"shader(
#version 330 core
in vec2 f_uv;
out vec4 out_color;
uniform sampler2D Tex;
void main()
{
    out_color = vec4(texture2D(Tex, f_uv).rgb, 1.0);
}
)shader";

void glfw_error(int const error, const char* description)
{
#if defined(_WIN32)
    __debugbreak();
#endif
    std::cerr << "GLFW error: " << error << "\n" << description << "\n";
}

void glfw_mouse_button(GLFWwindow* wnd, int const button, int const action, int modifiers)
{
    auto game = static_cast<Game*>(glfwGetWindowUserPointer(wnd));
    switch (action)
    {
    case GLFW_PRESS:
        game->on_mouse_down(static_cast<MouseButton>(button), static_cast<Modifiers>(modifiers));
        break;
    case GLFW_RELEASE:
        game->on_mouse_up(static_cast<MouseButton>(button), static_cast<Modifiers>(modifiers));
        break;
    default:
        std::abort();
    }
}

void glfw_mouse_move(GLFWwindow* wnd, double const mx, double const my);
void glfw_text(GLFWwindow* wnd, unsigned int const character)
{
    auto game = static_cast<Game*>(glfwGetWindowUserPointer(wnd));
    game->on_text(static_cast<u32>(character));
}
void glfw_key(GLFWwindow* wnd, int const key, int const /*scancode*/, int const action, int const modifiers)
{
    auto game = static_cast<Game*>(glfwGetWindowUserPointer(wnd));
    switch (action)
    {
    case GLFW_PRESS:
        game->on_key_down(static_cast<Key>(key), static_cast<Modifiers>(modifiers), false);
        break;
    case GLFW_REPEAT:
        game->on_key_down(static_cast<Key>(key), static_cast<Modifiers>(modifiers), true);
        break;
    case GLFW_RELEASE:
        game->on_key_up(static_cast<Key>(key), static_cast<Modifiers>(modifiers));
        break;
    default:
        std::abort();
    }
}

void glfw_framebuffer_resize(GLFWwindow* wnd, int, int);

} // namespace

class Game::Impl
{
public:
    Impl(char const* const title, i32 const width, i32 const height, i32 const p_scale_power) :
        surface{ width, height, p_scale_power },
        scale_power{ p_scale_power }
    {
        if (p_scale_power < 0 || p_scale_power > 3)
        {
            throw std::out_of_range{ "scale_power must be in range [0, 3]" };
        }
        if (width << p_scale_power < 128 || height << p_scale_power < 128)
        {
            throw std::out_of_range{ "scaled width and height must be at least 128" };
        }
        if (glfwInit() != GLFW_TRUE)
        {
            throw std::runtime_error{ "could not load glfw " };
        }

        glfwSetErrorCallback(glfw_error);

        glfwWindowHint(GLFW_CLIENT_API, GLFW_OPENGL_API);
        glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
        glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
        glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GLFW_FALSE);
        glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

        glfwWindowHint(GLFW_VISIBLE, GLFW_FALSE);
        glfwWindowHint(GLFW_RESIZABLE, GLFW_FALSE);

        glfwWindowHint(GLFW_RED_BITS, 8);
        glfwWindowHint(GLFW_GREEN_BITS, 8);
        glfwWindowHint(GLFW_BLUE_BITS, 8);
        glfwWindowHint(GLFW_DEPTH_BITS, 0);
        glfwWindowHint(GLFW_STENCIL_BITS, 0);

        wnd = { glfwCreateWindow(width << scale_power, height << scale_power, title, nullptr, nullptr) };
        if (wnd == nullptr)
        {
            throw std::runtime_error{ "could not create glfw window" };
        }

        glfwMakeContextCurrent(wnd);
        glfwSwapInterval(1);
        glfwSetFramebufferSizeCallback(wnd, glfw_framebuffer_resize);

        // IO callbacks
        glfwSetMouseButtonCallback(wnd, glfw_mouse_button);
        glfwSetCursorPosCallback(wnd, glfw_mouse_move);
        glfwSetCharCallback(wnd, glfw_text);
        glfwSetKeyCallback(wnd, glfw_key);

        if (!gladLoadGLLoader(reinterpret_cast<GLADloadproc>(glfwGetProcAddress)))
        {
            throw std::runtime_error{ "failed to initialize glad" };
        }
    }

    Impl           (Impl const&) = delete;
    Impl& operator=(Impl const&) = delete;

    ~Impl() noexcept
    {
        if (wnd != nullptr)
        {
            glfwDestroyWindow(wnd);
            glfwTerminate();
        }
    }

    GLFWwindow* wnd;
    Surface surface;
    i32 scale_power;
    i32 last_mouse_x = 0;
    i32 last_mouse_y = 0;
    bool vsync       = true;
};

namespace
{

void glfw_mouse_move(GLFWwindow* wnd, double mx, double my)
{
    auto game = static_cast<Game*>(glfwGetWindowUserPointer(wnd));
    auto& _ = game->_;
    auto const w = _->surface.width();
    auto const h = _->surface.height();
    auto const x = static_cast<i32>(mx) >> _->scale_power;
    auto const y = h - 1 - (static_cast<i32>(my) >> _->scale_power);

    if (x < 0 || x >= w ||
        y < 0 || y >= h)
    { return; }

    if (x == _->last_mouse_x && y == _->last_mouse_y)
    { return; }

    _->last_mouse_x = x;
    _->last_mouse_y = y;
    game->on_mouse_move(x, y);
}

void glfw_framebuffer_resize(GLFWwindow* wnd, int const fw, int const fh)
{
    auto& _ = static_cast<Game*>(glfwGetWindowUserPointer(wnd))->_;
    auto const tw = _->surface.width()  << _->scale_power;
    auto const th = _->surface.height() << _->scale_power;
    glfwSetWindowSize(wnd, tw, th);
    glViewport(0, 0, fw, fh);
}

} // namespace

Game::~Game() noexcept = default;

bool Game::vsync() const noexcept
{
    return _->vsync;
}

void Game::vsync(bool on) noexcept
{
    _->vsync = on;
    glfwSwapInterval(on ? 1 : 0);
}

void Game::resize(i32 const width, i32 const height, i32 const scale_power)
{
    if (scale_power < 0 || scale_power > 3)
    {
        throw std::out_of_range{ "scale_power must be in range [0, 3]" };
    }
    if (width << scale_power < 128 || height << scale_power < 128)
    {
        throw std::out_of_range{ "scaled width and height must be at least 128" };
    }
    
    _->scale_power = scale_power;
    _->surface = Surface{ width, height };
    glfwSetWindowSize(_->wnd, width << scale_power, height << scale_power);
}

void Game::exit() noexcept
{
    glfwSetWindowShouldClose(_->wnd, true);
}

void Game::run(Game& game)
{
    GlShaderProgram blit_shader =
    {
        GlShader{ ShaderType::Vertex,   vert_shader },
        GlShader{ ShaderType::Fragment, frag_shader },
    };
    GlTexture texture{ TextureWrap::ClampToBorder, TextureWrap::ClampToBorder, TextureFilter::Nearest, TextureFilter::Nearest };

    GLuint vbo, vao;
    glGenBuffers(1, &vbo);

    glGenVertexArrays(1, &vao);
    glBindVertexArray(vao);

    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, sizeof(float) * 2, 0);

    glEnableVertexAttribArray(0);

    static const float vertices[] =
    {
        -1.0f, -1.0f,
         1.0f, -1.0f,
        -1.0f,  1.0f,
        -1.0f,  1.0f,
         1.0f, -1.0f,
         1.0f,  1.0f,
    };
    glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices, GL_STATIC_DRAW);
    glBindVertexArray(0);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    if (glGetError() != GL_NO_ERROR)
    {
        __debugbreak();
    }

    auto& surface = game._->surface;
    auto const wnd = game._->wnd;

    glfwShowWindow(wnd);

    auto frame_end = TimePoint::clock::now();
    TimeSpan delta_time{ 0 };

    glClearColor(0, 0, 0, 0);

    while (!glfwWindowShouldClose(wnd))
    {
        glfwPollEvents();
        game.update(surface, delta_time);
        
        texture.update(surface);

        glClear(GL_COLOR_BUFFER_BIT);
        glUseProgram(blit_shader.id());
        glBindVertexArray(vao);
        glActiveTexture(GL_TEXTURE0 + 0);
        glBindTexture(GL_TEXTURE_2D, texture.id());
        glDrawArrays(GL_TRIANGLES, 0, 6);
        glBindVertexArray(0);
        glUseProgram(0);

        glfwSwapBuffers(wnd);

        auto const current_time = TimePoint::clock::now();
        delta_time = current_time - frame_end;
        frame_end = current_time;
    }
}

Game::Game(char const* const title, i32 const width, i32 const height, i32 const scale_power) :
    _{ std::make_unique<Impl>(title, width, height, scale_power) }
{
    glfwSetWindowUserPointer(_->wnd, this);
}

} // namespace
