#include "texture.hpp"
#include <framework/surface.hpp>
#include <cassert>
#include <stdexcept>

namespace fw
{

GlTexture::GlTexture() noexcept :
    tex_id_(0)
{
}

GlTexture::GlTexture(TextureWrap wrap_u, TextureWrap wrap_v, TextureFilter filter_min, TextureFilter filter_mag)
{
    assert(glGetError() == GL_NO_ERROR);
    glGenTextures(1, &tex_id_);
    if (tex_id_ == 0 || glGetError() != GL_NO_ERROR)
    {
        throw std::runtime_error{"could not create texture"};
    }
    glBindTexture(GL_TEXTURE_2D, tex_id_);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, static_cast<GLint>(wrap_u));
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, static_cast<GLint>(wrap_v));
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, static_cast<GLint>(filter_min));
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, static_cast<GLint>(filter_mag));

    glBindTexture(GL_TEXTURE_2D, 0);
}

GlTexture::GlTexture(GlTexture&& rhs) noexcept :
    GlTexture{}
{
    swap(rhs);
}

GlTexture& GlTexture::operator=(GlTexture&& rhs) noexcept
{
    auto move = std::move(rhs);
    swap(move);
    return *this;
}

GlTexture::~GlTexture() noexcept
{
    if (tex_id_ != 0)
    {
        glDeleteTextures(1, &tex_id_);
    }
}

void GlTexture::update(Surface const& surface)
{
    if (tex_id_ == 0)
    {
        throw std::logic_error{ "cannot update a null texture" };
    }
    assert(glGetError() == GL_NO_ERROR);
    using DataType = std::remove_pointer_t<decltype(surface.data())>;
    static_assert(std::is_standard_layout_v<DataType>, "DataType must be standard layout.");
    static_assert(sizeof(DataType) == 4, "DataType must be 4 bytes in size.");
    glBindTexture(GL_TEXTURE_2D, tex_id_);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8,
        static_cast<GLsizei>(surface.width()),
        static_cast<GLsizei>(surface.height()),
        0,
        GL_BGRA,
        GL_UNSIGNED_INT_8_8_8_8,
        surface.data());
    glBindTexture(GL_TEXTURE_2D, 0);
    assert(glGetError() == GL_NO_ERROR);
}

} // namespace fw
