#include <framework/game.hpp>
#include <framework/surface.hpp>
#include <particles/particle_simulation.hpp>
#include <iostream>
#include <sstream>

using namespace fw;

class ParticleGame : public Game
{
public:
    ParticleGame() :
        Game{ "Particle Runner", 256, 256, 2 },
        particles_{ particles::create_simulation() }
    { }

    void update(Surface& screen, TimeSpan const delta_time) override
    {
        screen.clear({});
        particles_->update(delta_time);
        particles_->render(screen);

        // C++17 structured binding
        auto[selected_index, selected_type, total_types] = get_particle_system_info();
        std::stringstream menu;
        menu << "F5: reset LMB: spawn <"
            << (selected_index + 1) << "/" << total_types << "> " << selected_type;
        screen.draw_text(menu.str(), 10, 10, { 255, 0, 0 });
        screen.draw_text(std::string{ "FPS: " } +std::to_string(1.f / delta_time.count()), 10, 20, 0x0000ffff);

        screen[mx_][my_] = { 0, 255, 255 };
    }

    void on_key_down(Key const key, Modifiers const /*modifiers*/, bool const repeat) override
    {
        if (repeat) { return; }
        switch (key)
        {
        case Key::Escape:
            exit();
            break;
        case Key::Right:
        case Key::D:
            ++selected_particle_type_;
            break;
        case Key::Left:
        case Key::A:
            --selected_particle_type_;
            break;

        case Key::F5:
            particles_->reset();
            break;

        case Key::F9:
            vsync(!vsync());
            break;

        default: break;
        }
    }

    void on_mouse_move(i32 mx, i32 my) override
    {
        mx_ = mx;
        my_ = my;
    }
    void on_mouse_down(MouseButton button, Modifiers /*modifiers*/) override
    {
        if (button == MouseButton::Left)
        {
            particles_->spawn_particle_system(get_particle_system_info().selected_name, static_cast<f32>(mx_), static_cast<f32>(my_));
        }
    }

private:
    struct ParticleSystemInfo
    {
        i32 selected_index;
        std::string_view selected_name;
        usize count;
    };
    ParticleSystemInfo get_particle_system_info() const noexcept
    {
        auto& types = particles_->get_particle_types();
        auto const size = static_cast<i32>(types.size());
        auto const index = (selected_particle_type_ % size + size) % size;
        return { index, types[index], types.size() };
    }

    particles::ParticleSimulationPtr particles_;
    i32 selected_particle_type_{};
    i32 mx_{};
    i32 my_{};
};

int main()
{
    try
    {
        Game::run(ParticleGame{});
    }
    catch (std::exception const& ex)
    {
        std::cerr << "Error while running game:\n" << ex.what() << "\n";
        return -1;
    }
    return 0;
}
