#include <type_traits>
#include <stdexcept>
#include <vector>
#include <stdint.h>

static_assert(std::is_same_v<unsigned char, uint8_t>);
using Archive = std::vector<uint8_t>;
using Size = int_fast32_t;

// Handles all integral types
template <typename T>
auto primitive_serialize(Archive& out, T value)
    -> std::enable_if_t<std::is_integral_v<T>, Size>
{
    for (auto i = 0; i < sizeof(T); ++i)
    {
        out.push_back(static_cast<uint8_t>(value >> i * 8));
    }
    return sizeof(T);
}
template <typename T>
auto primitive_deserialize(Archive const& in, Size& index, T& value)
    -> std::enable_if_t<std::is_integral_v<T>>
{
    if (in.size() - index < sizeof(T))
    {
        throw std::range_error{ "read out of range" };
    }
    value = T{};
    for (auto i = 0; i < sizeof(T); ++i)
    {
        value |= (T{} | in[index]) << i * 8;
        ++index;
    }
}

// Handle floating point types
template <typename T>
auto primitive_serialize(Archive& out, T value)
    -> std::enable_if_t<std::is_floating_point_v<T>, Size>
{
    auto ptr = static_cast<uint8_t*>(static_cast<void*>(&value));
    // Copy all the bytes
    for (auto i = 0; i < sizeof(T); ++i, ++ptr)
    {
        out.push_back(*ptr);
    }
    return sizeof(T);
}
template <typename T>
auto primitive_deserialize(Archive const& in, Size& index, T& value)
    -> std::enable_if_t<std::is_floating_point_v<T>>
{
    if (in.size() - index < sizeof(T))
    {
        throw std::range_error{ "read out of range" };
    }
    auto ptr = static_cast<uint8_t*>(static_cast<void*>(&value));
    for (auto i = 0; i < sizeof(T); ++i, ++ptr)
    {
        *ptr = in[index];
        ++index;
    }
}

template <typename T>
struct PrimitiveSerialization
{
    static Size serialize(Archive& out, T const& value)
    {
        return primitive_serialize(out, value);
    }

    static void deserialize(Archive const& in, Size& index, T& value)
    {
        primitive_deserialize(in, index, value);
    }
};

struct Vertex { float x, y, z; float u, v; uint32_t color; };
struct Mesh
{
    std::vector<Vertex> vertices;
    std::vector<uint32_t> indices;
};

template <template <typename> typename Serializer = PrimitiveSerialization>
struct MeshSerializer
{
private:
    template <typename T>
    static void write(Archive& out, T const& value)
    {
        Serializer<T>::serialize(out, value);
    }

    template <typename T>
    static void read(Archive const& in, Size& index, T& value)
    {
        Serializer<T>::deserialize(in, index, value);
    }

public:
    static Archive serialize(Mesh const& mesh)
    {
        Archive out;
        write(out, static_cast<uint32_t>(mesh.vertices.size()));
        write(out, static_cast<uint32_t>(mesh.indices.size()));

        for (auto const& vertex : mesh.vertices)
        {
            write(out, vertex.x);
            write(out, vertex.y);
            write(out, vertex.z);
            write(out, vertex.u);
            write(out, vertex.v);
            write(out, vertex.color);
        }
        for (auto const& index : mesh.indices)
        {
            write(out, index);
        }

        return out;
    }

    static Mesh deserialize(Archive const& in)
    {
        Mesh mesh;
        Size index = 0;
        uint32_t vertex_count, index_count;
        read(in, index, vertex_count);
        read(in, index, index_count);

        mesh.vertices.resize(vertex_count);
        mesh.indices.resize(index_count);

        for (uint32_t i = 0; i < vertex_count; ++i)
        {
            auto& vertex = mesh.vertices[i];
            read(in, index, vertex.x);
            read(in, index, vertex.y);
            read(in, index, vertex.z);
            read(in, index, vertex.u);
            read(in, index, vertex.v);
            read(in, index, vertex.color);
        }
        for (uint32_t i = 0; i < index_count; ++i)
        {
            read(in, index, mesh.indices[i]);
        }

        if (index != static_cast<Size>(in.size()))
        {
            throw std::invalid_argument{ "incompatible archive" };
        }

        return mesh;
    }
};

static void example()
{
    Mesh const mesh = /* mesh data */
    {
        {
            { 0.f, 0.f, 0.f, 0.f, 0.f, 0xff000000 },
            { 1.f, 0.f, 0.f, 1.f, 0.f, 0xffff0000 },
            { 1.f, 1.f, 0.f, 1.f, 1.f, 0xffffff00 },
            { 0.f, 1.f, 0.f, 0.f, 1.f, 0xff00ff00 }
        },
        {
            0, 1, 2,
            1, 2, 3
        }
    };

    auto const buffer = MeshSerializer<>::serialize(mesh);

    auto const mesh_copy = MeshSerializer<>::deserialize(buffer);
}
