#include <type_traits>
#include <cassert>

template <int Mass, int Length, int Time>
class Dimensions
{
public:
    static constexpr bool is_scalar = Mass == 0 && Length == 0 && Time == 0;

    float value{};
    Dimensions() noexcept = default;

    // Explicit conversion from float
    constexpr explicit Dimensions(float value) :
        value{ value }
    {}
    // Implicit conversion to float, iif it's Scalar
    template <typename = std::enable_if_t<is_scalar>>
    constexpr operator float() const noexcept { return value; }
};
using Scalar = Dimensions<0, 0, 0>;

// Arithmetic operators
template <int M, int L, int T>
constexpr auto operator+(Dimensions<M, L, T> lhs, Dimensions<M, L, T> rhs) noexcept
{
    return lhs.value + rhs.value;
}

template <int M, int L, int T>
constexpr auto operator-(Dimensions<M, L, T> lhs, Dimensions<M, L, T> rhs) noexcept
{
    return lhs.value - rhs.value;
}

template <int LM, int LL, int LT, int RM, int RL, int RT>
constexpr auto operator*(Dimensions<LM, LL, LT> lhs, Dimensions<RM, RL, RT> rhs) noexcept
{
    return Dimensions<LM + RM, LL + RL, LT + RT>{ lhs.value * rhs.value };
}
template <int LM, int LL, int LT, int RM, int RL, int RT>
constexpr auto operator/(Dimensions<LM, LL, LT> lhs, Dimensions<RM, RL, RT> rhs) noexcept
{
    return Dimensions<LM - RM, LL - RL, LT - RT>{ lhs.value / rhs.value };
}

template <int M, int L, int T>
constexpr auto operator+=(Dimensions<M, L, T>& lhs, Dimensions<M, L, T> rhs)
noexcept -> Dimensions<M, L, T>&
{
    lhs.value += rhs.value;
    return lhs;
}
template <int M, int L, int T>
constexpr auto operator-=(Dimensions<M, L, T>& lhs, Dimensions<M, L, T> rhs)
noexcept -> Dimensions<M, L, T>&
{
    lhs.value -= rhs.value;
    return lhs;
}

template <int M, int L, int T, typename S>
    constexpr auto operator*=(Dimensions<M, L, T>& lhs, S const& rhs)
    noexcept -> std::enable_if_t<std::is_convertible_v<S, float>, Dimensions<M, L, T>&>
{
    lhs.value *= static_cast<float>(rhs);
    return lhs;
}
template <int M, int L, int T, typename S>
    constexpr auto operator/=(Dimensions<M, L, T>& lhs, S const& rhs)
    noexcept -> std::enable_if_t<std::is_convertible_v<S, float>, Dimensions<M, L, T>&>
{
    lhs.value /= static_cast<float>(rhs);
    return lhs;
}

template <int M, int L, int T>
constexpr bool operator==(Dimensions<M, L, T> lhs, Dimensions<M, L, T> rhs) noexcept
{
    return lhs.value == rhs.value;
}
template <int M, int L, int T>
constexpr bool operator!=(Dimensions<M, L, T> lhs, Dimensions<M, L, T> rhs) noexcept
{
    return lhs.value != rhs.value;
}

using Mass   = Dimensions< 1, 0, 0>;
using Length = Dimensions< 0, 1, 0>;
using Time   = Dimensions< 0, 0, 1>;

using Velocity     = Dimensions< 0,  1, -1>; // Length   / Time
using Acceleration = Dimensions< 0,  1, -2>; // Velocity / Time
using Force        = Dimensions< 1,  1, -2>; // Mass     * Acceleration
using Frequency    = Dimensions< 0,  0, -1>; // 1        / Time
using Pressure     = Dimensions< 1, -1, -2>; // Force    / Length^2
using Energy       = Dimensions< 1,  2, -2>; // Force    * Length
using Power        = Dimensions< 1,  2, -3>; // Energy   / Time

// Alternative purposes with the same units
using Weight = Force;
using Stress = Pressure;
using Work = Energy;
using Heat = Energy;

static_assert(std::is_convertible_v<Scalar, float>);

static void example()
{
    constexpr Velocity speed = Length{ 10 } / Time{ 2 };
    static_assert(speed == Velocity{ 5 });

    constexpr Weight weight = Mass{ 2 } * speed / Time { 4 };
    static_assert(weight == Force{ 2.5f });

    constexpr Work work = Length{ 10 } * weight;
    static_assert(work == Work{ 25 });

    Acceleration gravity{ 1.f };
    gravity *= 9.78f;
    assert(gravity == Acceleration{ 9.78f });
    gravity /= 2.f;
    assert(gravity == Acceleration{ 9.78f / 2.f });
    
    Length length{ 1.f };
    length += Length{ 10.f };
    length -= Length{ 1.f };
    assert(length == Length{ 10.f });

    // length += 10.f;
    // length *= Length{ 10.f };
}
