#pragma once

#include "type_list.hpp"
#include "particle_input.hpp"

/*
 * Each property of a particle consists of 2 primary parts.
 * 
 * 1. The data class, this holds the data requires for the
 *    particle to be simulated, as well as the update logic.
 *    
 * 2. The property class, this holds the data for
 *    initializing instances of the data class.
 *    
 * Because MSVC does not yet implement class template type
 * deduction, there is also a 3rd part, which is:
 * 
 * 3. Factory function for creating property class.
 * 
 * Each data class can require dependencies on other data
 * classes. It does this through a type alias Dependencies.
 * For example, VelocityData specifies:
 * using Dependencies = TypeList<PositionData>;
 * 
 * Each data class optionally requires updates, if it does
 * it should specify a function:
 * void update(..., float delta_seconds) noexecpt;
 * The ... should be a list of references to the dependencies.
 */

namespace particles::detail
{

/*
 * 1. Data classes
 */
struct LifetimeData
{
    using Dependencies = TypeList<>;

    float life_progress_ = 0.f;
    float life_rate_     = 0.f; // 1 / total_lifetime

    void update(float delta_seconds) noexcept
    {
        life_progress_ += life_rate_ * delta_seconds;
    }
};

struct PositionData
{
    using Dependencies = TypeList<>;

    float x{};
    float y{};
};
struct VelocityData
{
    using Dependencies = TypeList<PositionData>;

    float x{};
    float y{};

    void update(PositionData& pos, float delta_seconds) noexcept
    {
        pos.x += x * delta_seconds;
        pos.y += y * delta_seconds;
    }
};
struct AccelerationData
{
    using Dependencies = TypeList<VelocityData>;

    float x{};
    float y{};

    void update(VelocityData& vel, float delta_seconds) noexcept
    {
        vel.x += x * delta_seconds;
        vel.y += y * delta_seconds;
    }
};
struct DragData
{
    using Dependencies = TypeList<VelocityData>;

    float x{};
    float y{};

    void update(VelocityData& vel, float delta_seconds) noexcept
    {
        vel.x *= std::max(0.f, 1.f - delta_seconds * x);
        vel.y *= std::max(0.f, 1.f - delta_seconds * y);
    }
};

/*
 * 2. Property classes
 */
template <typename T>
struct LifetimeImpl
{
    using Data = LifetimeData;
    constexpr LifetimeImpl() noexcept = default;
    T value{};

    template <typename D>
    void init(D& data, RandomEngine& random) const
    {
        auto& lt = static_cast<LifetimeData&>(data);
        auto generated = value.generate(random);
        if (generated <= 0.f)
        {
            throw std::invalid_argument{ "particle type with negative lifetime" };
        }
        lt.life_rate_ = 1.f / generated;
    }
};

/*
 * Curiously Recurring Template pattern.
 * This time using Template Template parameters.
 */
template <typename Derived>
struct VectorImplBase;
template <typename TX, typename TY, template <typename, typename> typename T>
struct VectorImplBase<T<TX, TY>>
{
    using Derived = T<TX, TY>;

    constexpr VectorImplBase(TX px, TY py) noexcept :
        x{ px },
        y{ py }
    {}
    TX const x{};
    TY const y{};
    
    template <typename D>
    void init(D& data, RandomEngine& random) const
    {
        auto& obj = static_cast<typename Derived::Data&>(data);
        obj.x = x.generate(random);
        obj.y = y.generate(random);
    }
};
template <template <typename, typename> typename T>
struct VectorImplBase<T<Polar, Polar>>
{
    using Derived = T<Polar, Polar>;

    constexpr VectorImplBase(Polar ppolar, Polar) noexcept :
        polar{ ppolar }
    {}

    Polar polar;

    template <typename D>
    void init(D& data, RandomEngine& random) const
    {
        auto& obj = static_cast<typename Derived::Data&>(data);
        auto[x, y] = polar.generate(random);
        obj.x = x;
        obj.y = y;
    }
};

template <typename TX, typename TY>
struct PositionImpl final : VectorImplBase<PositionImpl<TX, TY>>
{
    using Data = PositionData;
    constexpr PositionImpl(TX px, TY py) noexcept :
        VectorImplBase<PositionImpl>{ px, py }
    {}
};

template <typename TX, typename TY>
struct VelocityImpl final : VectorImplBase<VelocityImpl<TX, TY>>
{
    using Data = VelocityData;
    constexpr VelocityImpl(TX px, TY py) noexcept :
    VectorImplBase<VelocityImpl>{ px, py }
    {}
};

template <typename TX, typename TY>
struct AccelerationImpl final : VectorImplBase<AccelerationImpl<TX, TY>>
{
    using Data = AccelerationData;
    constexpr AccelerationImpl(TX px, TY py) noexcept :
    VectorImplBase<AccelerationImpl>{ px, py }
    {}
};

template <typename TX, typename TY>
struct DragImpl final : VectorImplBase<DragImpl<TX, TY>>
{
    using Data = DragData;
    constexpr DragImpl(TX px, TY py) noexcept :
    VectorImplBase<DragImpl>{ px, py }
    {}
};
    
} // particles::detail

/*
 * 3. Factory functions for property classes.
 */
namespace particles
{

template <typename T>
constexpr auto Lifetime(T value)
    noexcept -> detail::LifetimeImpl<T>
{
    return { value };
}

template <typename T>
constexpr auto Position(T value)
    noexcept -> detail::PositionImpl<T, T>
{
    return { value, value };
}
template <typename TX, typename TY>
constexpr auto Position(TX x, TY y)
    noexcept -> detail::PositionImpl<TX, TY>
{
    return { x, y };
}

template <typename T>
constexpr auto Velocity(T value)
noexcept -> detail::VelocityImpl<T, T>
{
    return { value, value };
}
template <typename TX, typename TY>
constexpr auto Velocity(TX x, TY y)
noexcept -> detail::VelocityImpl<TX, TY>
{
    return { x, y };
}

template <typename T>
constexpr auto Acceleration(T value)
noexcept -> detail::AccelerationImpl<T, T>
{
    return { value, value };
}
template <typename TX, typename TY>
constexpr auto Acceleration(TX x, TY y)
noexcept -> detail::AccelerationImpl<TX, TY>
{
    return { x, y };
}

template <typename T>
constexpr auto Drag(T value)
    noexcept -> detail::DragImpl<T, T>
{
    return { value, value };
}
template <typename TX, typename TY>
constexpr auto Drag(TX x, TY y)
    noexcept -> detail::DragImpl<TX, TY>
{
    return { x, y };
}

} // namespace particles
