#pragma once

#include "particle_data.hpp"
#include "particle_random.hpp"
#include <type_traits>
#include <array>
#include <tuple>

namespace particles::detail
{
    
/*
 * MakeDataList is a utility type alias that generates
 * a TypeList for data classes including their dependencies.
 *
 * It only contains unique types.
 *
 * Example:
 * MakeDataList<VelocityData> ==> TypeList<VelocityData, PositionData>
 * MakeDataList<PositionData> ==> TypeList<PositionData>
 * MakeDataList<PositionData, VelocityData> ==> TypeList<PositionData, VelocityData>
 */

template <typename, typename>
struct MakeDataListImpl;
template <typename... Result>
struct MakeDataListImpl<TypeList<Result...>, TypeList<>>
{
    using Type = UniqueTypeList<Result...>;
};
template <typename... Result, typename First, typename... Rest>
struct MakeDataListImpl<TypeList<Result...>, TypeList<First, Rest...>>
{
    using Type = typename MakeDataListImpl<
        ConcatTypeList<TypeList<Result..., First>, typename First::Dependencies>,
        TypeList<Rest...>
    >::Type;
};

template <typename... Data>
using MakeDataList = typename MakeDataListImpl<
    TypeList<>, TypeList<Data...>
>::Type;

template <typename T, typename TypeList, typename = void>
constexpr bool has_update_function_helper = false;
template <typename T, typename... Dependencies>
constexpr bool has_update_function_helper<T, TypeList<Dependencies...>,
    decltype(static_cast<T*>(nullptr)->update(*static_cast<Dependencies*>(nullptr)..., 0.f))> =
    true;

template <typename T, typename = void>
constexpr bool has_update_function = has_update_function_helper<T, typename T::Dependencies>;

/*
* invoke_update uses SFINAE to check if the data type has
* requires_update set to true before trying to call update(...)
*/
template <typename DataType, typename Particle, typename... Deps>
auto invoke_update(Particle& data, float delta_seconds, TypeList<Deps...>)
    noexcept -> std::enable_if_t<has_update_function<DataType>>
{
    static_cast<DataType&>(data).update(static_cast<Deps&>(data)..., delta_seconds);
}
template <typename DataType, typename Particle, typename... Deps>
auto invoke_update(Particle&, float, TypeList<Deps...>)
    noexcept -> std::enable_if_t<!has_update_function<DataType>>
{
}

/*
 * ParticleData is the actual data class containing all the
 * required fields for the particle to be simulated.
 * 
 * It is created by ParticleType::create(), and inherits
 * from classes such as LifetimeData.
 * 
 * It has 4 member functions:
 *   update(float delta_seconds)
 *   is_alive() const -> bool
 *   position() const -> std::array<float, 2>
 *   position(f32 x, f32 y)
 */
template <typename>
struct ParticleData;
template <typename... Data>
struct ParticleData<TypeList<Data...>> : Data...
{
    void update(float delta_seconds) noexcept
    {
        // Calls invoke_update for each Data class
        auto _ = { (
            invoke_update<Data>(*this, delta_seconds, typename Data::Dependencies{})
        , 0)... };
    }

    bool is_alive() const noexcept
    {
        return static_cast<LifetimeData const&>(*this).life_progress_ < 1.f;
    }

    void position(float x, float y) noexcept
    {
        auto& pos = static_cast<PositionData&>(*this);
        pos.x = x;
        pos.y = y;
    }
    std::array<float, 2> position() const noexcept
    {
        auto& pos = static_cast<PositionData const&>(*this);
        return { pos.x, pos.y };
    }
};

/*
 * ParticleType is a factory class for ParticleData.
 * Its only public API is:
 *   create(RandomEngine&) -> Particle
 * 
 * ParticleType is a compile-time class, and only stores
 * the information necessary to create particles.
 * 
 * It is created by CreateParticleType, example:
 * constexpr auto particle_type = CreateParticleType(
 *     Lifetime(Constant{ 1.f }),
 *     Position(Constant{ 0.f }, Constant{ 0.f })
 * );
 */
template <typename... Properties>
struct ParticleType
{
private:
    using DataList = MakeDataList<typename Properties::Data...>;

    // All particles must have a lifetime and position
    static_assert(
        detail::ListContainsType<LifetimeData, DataList>,
        "particle must have a lifetime");
    static_assert(
        detail::ListContainsType<PositionData, DataList>,
        "particle must have a position");

    std::tuple<Properties...> properties_;

public:
    constexpr explicit ParticleType(Properties... properties) :
        properties_{ std::move(properties)... }
    {
    }

    using Particle = ParticleData<DataList>;
    Particle create(RandomEngine& random, float x, float y) const
    {
        auto particle = create_impl(random, std::make_index_sequence<sizeof...(Properties)>{});
        auto& pos = static_cast<PositionData&>(particle);
        pos.x += x;
        pos.y += y;
        return particle;
    }

private:
    // Uses std::index_sequence to deduce a list of indices
    // This pack is then expanded into std::get<N> to access
    // each element stored in the properties_ tuple.
    template <size_t... Indices>
    Particle create_impl(RandomEngine& random, std::index_sequence<Indices...>) const
    {
        Particle data{};
        // Calls init(data, random) on each property in properties_
        auto _ = { (static_cast<void>(
            std::get<Indices>(properties_).init(data, random)
        ), 0)... };
        return data;
    }
};

}

namespace particles
{
    
/*
 * Creates a particle type. Pass it various data properties.
 * See particle_data.hpp for the available properties.
 */
template <typename... Properties>
constexpr auto CreateParticleType(Properties... properties)
{
    return detail::ParticleType<Properties...>
    {
        std::move(properties)...
    };
}

} // namespace particles
